/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this String except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.crypto;

import de.gematik.ti.sgd.exceptions.generic.CertificateValidationException;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.StringUtils;
import org.bouncycastle.asn1.*;
import org.bouncycastle.asn1.x509.Certificate;
import org.bouncycastle.cert.ocsp.BasicOCSPResp;
import org.bouncycastle.cert.ocsp.OCSPResp;
import org.bouncycastle.crypto.digests.SHA256Digest;
import org.bouncycastle.crypto.generators.HKDFBytesGenerator;
import org.bouncycastle.crypto.params.HKDFParameters;
import org.bouncycastle.jcajce.provider.asymmetric.ec.BCECPublicKey;
import org.bouncycastle.jce.ECNamedCurveTable;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.jce.spec.ECNamedCurveParameterSpec;
import org.bouncycastle.jce.spec.ECNamedCurveSpec;
import org.bouncycastle.operator.jcajce.JcaContentVerifierProviderBuilder;

import javax.crypto.Cipher;
import javax.crypto.KeyAgreement;
import javax.crypto.SecretKey;
import javax.crypto.spec.GCMParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import javax.xml.bind.DatatypeConverter;
import java.io.ByteArrayInputStream;
import java.io.InputStream;
import java.math.BigInteger;
import java.security.*;
import java.security.cert.CertificateFactory;
import java.security.cert.CertificateParsingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.ECPublicKey;
import java.security.spec.ECParameterSpec;
import java.security.spec.ECPoint;
import java.security.spec.ECPublicKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.concurrent.ThreadLocalRandom;
import java.util.function.Function;

public class SgdCryptoUtilitiesImplementation extends SgdCryptoUtilities {
    private static final String BRAINPOOL_P_256_R1 = "brainpoolp256r1";
    private static final String OCSP_SIGNER_OID = "1.3.6.1.5.5.7.3.9";
    private static final String ADMISSION_EXTENSION_OID = "1.3.36.8.3.3";
    private static final int GCM_IV_LENGTH = 12;
    private static final int GCM_TAG_LENGTH = 16;
    private static final String FIND_LEADING_ZEROS = "^0+";
    private final SecureRandom SECURE_RANDOM;
    private final BouncyCastleProvider provider;

    public SgdCryptoUtilitiesImplementation() {
        provider = new BouncyCastleProvider();
        Security.addProvider(provider);

        try {
            SECURE_RANDOM = SecureRandom.getInstance("SHA1PRNG");
        } catch (NoSuchAlgorithmException e) {
            throw new SgdCryptoStartupException("Crypto initialization failed!");
        }
        setInstance(this);
    }

    private static boolean isOcspSignerCertificate(X509Certificate x509Certificate) {
        try {
            if (x509Certificate == null || x509Certificate.getExtendedKeyUsage() == null) {
                return false;
            }
            return x509Certificate.getExtendedKeyUsage().stream()
                    .filter(oidStr -> OCSP_SIGNER_OID.equals(oidStr))
                    .findAny().isPresent();
        } catch (CertificateParsingException e) {
            return false;
        }
    }

    private static Function<X509Certificate, Boolean> isOcspResponseSignedWithGivenCertificate(BasicOCSPResp ocspResponse) {
        return certificate -> {
            try {
                return ocspResponse.isSignatureValid(new JcaContentVerifierProviderBuilder().build(certificate));
            } catch (Exception e) {
                return false;
            }
        };
    }

    @Override
    public Optional<String> extractTelematikIdFromCertificate(X509Certificate certificate) {
        try {
            final ASN1Encodable parsedValue = Certificate.getInstance(certificate.getEncoded())
                    .getTBSCertificate().getExtensions().getExtension(new ASN1ObjectIdentifier(ADMISSION_EXTENSION_OID))
                    .getParsedValue();

            DLSequence a = (DLSequence) parsedValue;
            DLSequence b = null;
            final Iterator<ASN1Encodable> iterator = a.iterator();
            while(iterator.hasNext()) {
                final ASN1Encodable next = iterator.next();
                if (next instanceof DLSequence) {
                    b = (DLSequence) next;
                }
            }
            DLSequence c = (DLSequence) b.getObjectAt(0);
            DLSequence d = (DLSequence) c.getObjectAt(0);
            DLSequence e = (DLSequence) d.getObjectAt(0);
            DERPrintableString string = (DERPrintableString) e.getObjectAt(2);

            return Optional.of(string.getString());
        } catch (Exception e) {
            return Optional.empty();
        }
    }

    private void performOperationSafe(UnsafeOperation operation, String errorMessage) {
        try {
            operation.perform();
        } catch (Exception e) {
            throw new SgdCryptoException(errorMessage, e);
        }
    }

    private <T> T performOperationSafe(UnsafeSupplier<T> supplier, String errorMessage) {
        T result;
        try {
            result = supplier.get();
        } catch (Exception e) {
            throw new SgdCryptoException(errorMessage, e);
        }

        if (result == null) {
            throw new SgdCryptoException(errorMessage);
        }
        return result;
    }

    @Override
    public X509Certificate loadX509CertificateFromP12File(String file, String p12Password) {
        return performOperationSafe(() -> {
            KeyStore p12 = KeyStore.getInstance("pkcs12", provider);
            p12.load(ClasspathFileUtil.getInputStreamForPath(file), p12Password.toCharArray());
            Enumeration e = p12.aliases();
            if (e.hasMoreElements()) {
                String alias = (String) e.nextElement();
                return (X509Certificate) p12.getCertificate(alias);
            } else {
                throw new SgdCryptoException("Could not find any aliases in p12 file '" + file + "'");
            }
        }, "Error while loading certificate from file " + file);
    }

    @Override
    public byte[] convertB64StringToBytes(String b64String) {
        return DatatypeConverter.parseBase64Binary(b64String);
    }

    @Override
    public String convertBytesToB64String(byte[] encoded) {
        return DatatypeConverter.printBase64Binary(encoded);
    }

    @Override
    public X509Certificate loadX509CertificateFromCrtFile(String crtFile) {
        return performOperationSafe(() -> {
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509", provider);
            return (X509Certificate) certFactory.generateCertificate(ClasspathFileUtil.getInputStreamForPath(crtFile));
        }, "Error while loading certificate from '" + crtFile + "'");
    }

    @Override
    public PrivateKey loadPrivateKeyFromP12File(String p12File, String p12Password, String entryPassword) {
        return performOperationSafe(() -> {
            KeyStore p12 = KeyStore.getInstance("pkcs12", provider);
            p12.load(ClasspathFileUtil.getInputStreamForPath(p12File), p12Password.toCharArray());
            PrivateKey privateKey = null;
            Enumeration e = p12.aliases();
            while (e.hasMoreElements()) {
                String alias = (String) e.nextElement();
                privateKey = (PrivateKey) p12.getKey(alias, entryPassword.toCharArray());
            }

            return privateKey;
        }, "Error while loading private key from '" + p12File + "'");
    }

    @Override
    public byte[] createSignatureUsingEcdsa(byte[] toBeSignedData, PrivateKey signerKey) {
        return performOperationSafe(() -> {
            Signature signer = Signature.getInstance("SHA256withECDSA", provider);
            signer.initSign(signerKey);
            signer.update(toBeSignedData);
            return signer.sign();
        }, "Error while creating Signature");
    }

    @Override
    public KeyPair loadKeypairFromP12File(String p12File, String p12Password, String entryPassword) {
        PrivateKey privateKey = loadPrivateKeyFromP12File(p12File, p12Password, entryPassword);
        PublicKey publicKey = loadX509CertificateFromP12File(p12File, p12Password).getPublicKey();
        return new KeyPair(publicKey, privateKey);
    }

    @Override
    public KeyPair generateEccKeyPair() {
        return generateEccKeyPair(BRAINPOOL_P_256_R1);
    }

    @Override
    public KeyPair generateEccKeyPair(String curveName) {
        return performOperationSafe(() -> {
            KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", provider);
            ECNamedCurveParameterSpec dp = ECNamedCurveTable.getParameterSpec(curveName);
            kpg.initialize(dp, SECURE_RANDOM);
            return kpg.generateKeyPair();
        }, "Error while generating ECC KeyPair");
    }

    @Override
    public byte[] parseHexStringToByteArray(String hexString) {
        return DatatypeConverter.parseHexBinary(hexString);
    }

    @Override
    public String convertByteArrayToHexString(byte[] byteArray) {
        return DatatypeConverter.printHexBinary(byteArray);
    }

    @Override
    public boolean doesCertificateContainAdmissionsExtensionOid(X509Certificate certificate, String admissionsOid) {
        byte[] extensionValue = certificate.getExtensionValue(ADMISSION_EXTENSION_OID);
        if (extensionValue == null) {
            return false;
        }

        String admissionsExtensionString =
                ASN1Sequence.getInstance(
                        Arrays.copyOfRange(extensionValue, 2, extensionValue.length)) // cut away the octet-String T and L
                        .toString();

        return admissionsExtensionString.contains(admissionsOid);
    }

    @Override
    public boolean isSgdHsmCertificateInTsl(X509Certificate certificate) {
        // No check
        return true;
    }

    @Override
    public String extractPublicPointCoordinatesFromPublicKey(PublicKey publicKey) {
        return performOperationSafe(() -> {
            X509EncodedKeySpec spec = new X509EncodedKeySpec(publicKey.getEncoded());
            KeyFactory kf = KeyFactory.getInstance("EC", provider);
            ECPublicKey ecPublicKey = (ECPublicKey) kf.generatePublic(spec);
            String hexValueAffineX = convertToLowerCaseHexWith64ByteLength(ecPublicKey.getW().getAffineX());
            String hexValueAffineY = convertToLowerCaseHexWith64ByteLength(ecPublicKey.getW().getAffineY());
            return "0x" + hexValueAffineX.replaceAll(FIND_LEADING_ZEROS, "")
                    + " 0x" + hexValueAffineY.replaceAll(FIND_LEADING_ZEROS, "");
        }, "Error while extracting X/Y from public Key!");
    }

    @Override
    public String convertToLowerCaseHexWith64ByteLength(BigInteger value) {
        return convertToLowerCaseHexWith64ByteLength(value.toByteArray());
    }

    @Override
    public String convertToLowerCaseHexWith64ByteLength(byte[] rawBytes) {
        final byte[] byteArrayWithLimitedLength = ArrayUtils.subarray(rawBytes, rawBytes.length - 32, rawBytes.length);

        return convertByteArrayToHexString(byteArrayWithLimitedLength).toLowerCase();
    }

    @Override
    public byte[] performHkdf(byte[] ikm, byte[] info, int length) {
        return performOperationSafe(() -> {
            HKDFBytesGenerator hkdf = new HKDFBytesGenerator(new SHA256Digest());
            hkdf.init(new HKDFParameters(ikm, null, info));
            byte[] okm = new byte[length / 8];
            hkdf.generateBytes(okm, 0, length / 8);
            return okm;
        }, "Error while performing HKDF");
    }

    @Override
    public void verifySignatureAndThrowExceptionWhenFail(byte[] toBeSignedBytes, PublicKey publicKey, byte[] signatureBytes) throws SignatureException {
        if (publicKey instanceof BCECPublicKey) {
            verifyEcdsaSignatureAndThrowExceptionWhenFail(toBeSignedBytes, publicKey, signatureBytes);
        } else {
            verifyRsaSsaPssSignatureAndThrowExceptionWhenFail(toBeSignedBytes, publicKey, signatureBytes);
        }
    }

    @Override
    public void verifyEcdsaSignatureAndThrowExceptionWhenFail(byte[] toBeSignedBytes, PublicKey publicKey, byte[] signatureBytes) throws SignatureException {
        if (!performOperationSafe(() -> {
            Signature signer = Signature.getInstance("SHA256withECDSA", provider);
            signer.initVerify(publicKey);
            signer.update(toBeSignedBytes);
            return signer.verify(signatureBytes);
        }, "Error while validating signature")) {
            throw new SignatureException("Signature is not valid");
        }
    }

    @Override
    public void verifyRsaSsaPssSignatureAndThrowExceptionWhenFail(byte[] toBeSignedBytes, PublicKey publicKey, byte[] signatureBytes) throws SignatureException {
        if (!performOperationSafe(() -> {
            Signature signer = Signature.getInstance("SHA256withRSAAndMGF1", provider);
            signer.initVerify(publicKey);
            signer.update(toBeSignedBytes);
            return signer.verify(signatureBytes);
        }, "Error while validating signature")) {
            throw new SignatureException("Signature is not valid");
        }
    }

    @Override
    public X509Certificate parseX509CertificateFromB64String(String certificate) {
        return parseX509CertificateFromBytes(convertB64StringToBytes(certificate));
    }

    @Override
    public void verifyOcspSignatureAndThrowExceptionWhenFail(byte[] ocspResponseRawBytes, List<X509Certificate> possibleOcspSignerCertificates) {
        performOperationSafe(() -> {
            final OCSPResp ocspResponse = new OCSPResp(ocspResponseRawBytes);
            BasicOCSPResp basicResponse = (BasicOCSPResp) ocspResponse.getResponseObject();
            assertThatOcspResponseSignatureIsValid(basicResponse, possibleOcspSignerCertificates);
            ZonedDateTime producedAt = ZonedDateTime.ofInstant(basicResponse.getProducedAt().toInstant(), ZoneId.of("UTC"));
            if (producedAt.isBefore(ZonedDateTime.now().minusHours(4))) {
                throw new SgdCryptoException("OCSP-Response is older than four hours.");
            }
        }, "Error while loading certificate from bytes!");
    }

    @Override
    public X509Certificate parseX509CertificateFromBytes(byte[] sgdCertificateRawBytes) {
        return performOperationSafe(() -> {
            CertificateFactory certFactory = CertificateFactory.getInstance("X.509", provider);
            InputStream in = new ByteArrayInputStream(sgdCertificateRawBytes);
            final X509Certificate x509Certificate = (X509Certificate) certFactory.generateCertificate(in);
            if (x509Certificate == null) {
                throw new SgdCryptoException("Error while loading certificate!");
            }
            return x509Certificate;
        }, "Error while loading certificate from bytes!");
    }

    private void assertThatOcspResponseSignatureIsValid(BasicOCSPResp ocspResponse, List<X509Certificate> possibleSigners) throws CertificateValidationException {
        boolean responseIsSignedBySignerOnList = possibleSigners.stream()
                .filter(x509Certificate -> isOcspSignerCertificate(x509Certificate))
                .map(isOcspResponseSignedWithGivenCertificate(ocspResponse))
                .filter(value -> value)
                .findAny().isPresent();
        if (!responseIsSignedBySignerOnList) {
            throw new CertificateValidationException("Could not verify given OCSP Response!");
        }
    }

    @Override
    public byte[] deriveKeyAndEncrypt(PrivateKey senderPrivate, PublicKey receiverPublic, byte[] payload) {
        return performOperationSafe(() -> {
            byte[] aesKey = deriveAesEncryptionKey(senderPrivate, receiverPublic);
            return encryptUsingAesGcm(payload, aesKey, null);
        }, "Error during ECIES encryption");
    }

    @Override
    public byte[] deriveAesEncryptionKey(PrivateKey selfPrivateKey, PublicKey otherPublicKey) {
        return performOperationSafe(() -> {
            KeyAgreement ka = KeyAgreement.getInstance("ECDHC");
            ka.init(selfPrivateKey);
            ka.doPhase(otherPublicKey, true);
            byte[] sharedSecret = ka.generateSecret();

            HKDFBytesGenerator hkdf = new HKDFBytesGenerator(new SHA256Digest());
            hkdf.init(HKDFParameters.defaultParameters(sharedSecret));
            byte[] aesKey = new byte[32];
            hkdf.generateBytes(aesKey, 0, 32);

            return aesKey;
        }, "Error during ECIES Key Derivation");
    }

    @Override
    public byte[] encryptUsingAesGcm(byte[] input, byte[] key, byte[] associatedData) {
        return performOperationSafe(() -> {
            byte[] iv = generateIV();
            byte[] cipherTextPlusTag;
            SecretKey secretKey = new SecretKeySpec(key, "AES");
            Cipher cipher = Cipher.getInstance("AES/GCM/NoPadding", provider);

            cipher.init(Cipher.ENCRYPT_MODE, secretKey, new GCMParameterSpec(GCM_TAG_LENGTH * 8, iv));

            if (associatedData != null) {
                cipher.updateAAD(associatedData);
            }

            cipherTextPlusTag = cipher.doFinal(input);

            byte[] encMessage = Arrays.copyOf(iv, GCM_IV_LENGTH + cipherTextPlusTag.length);
            System.arraycopy(cipherTextPlusTag, 0, encMessage, GCM_IV_LENGTH, cipherTextPlusTag.length);

            return encMessage;
        }, "Error during AES-GCM encryption");
    }

    @Override
    public byte[] decryptUsingAesGcm(byte[] encMessage, byte[] key, byte[] associatedData) {
        return performOperationSafe(() -> {
            SecretKey secretKey = new SecretKeySpec(key, "AES");
            byte[] plainText = null;
            byte[] iv = Arrays.copyOfRange(encMessage, 0, GCM_IV_LENGTH);
            byte[] cipherText = Arrays.copyOfRange(encMessage, GCM_IV_LENGTH, encMessage.length);
            Cipher cipher = Cipher.getInstance("AES/GCM/NoPadding", provider);
            cipher.init(Cipher.DECRYPT_MODE, secretKey, new GCMParameterSpec(GCM_TAG_LENGTH * 8, iv));

            if (associatedData != null) {
                cipher.updateAAD(associatedData);
            }

            plainText = cipher.doFinal(cipherText);

            return plainText;
        }, "Error during AES-GCM decryption");
    }

    @Override
    public PublicKey extractPublicKeyFromPublicPointCoordinates(String xString, String yString, String curveName) {
        return performOperationSafe(() -> {
            final String xCoordinate = StringUtils.leftPad(xString.replace("0x", ""), 64, '0');
            final String yCoordinate = StringUtils.leftPad(yString.replace("0x", ""), 64, '0');
            byte[] x = DatatypeConverter.parseHexBinary(xCoordinate);
            byte[] y = DatatypeConverter.parseHexBinary(yCoordinate);

            KeyFactory eckf = KeyFactory.getInstance("EC", provider);
            ECPoint point = new ECPoint(new BigInteger(1, x), new BigInteger(1, y));
            ECNamedCurveParameterSpec parameterSpec = ECNamedCurveTable.getParameterSpec(curveName);
            ECParameterSpec spec = new ECNamedCurveSpec(curveName, parameterSpec.getCurve(), parameterSpec.getG(), parameterSpec.getN(), parameterSpec.getH(), parameterSpec.getSeed());
            ECPublicKey ecPublicKey = (ECPublicKey) eckf.generatePublic(new ECPublicKeySpec(point, spec));
            return ecPublicKey;
        }, "Error during Public-Key extraction");
    }

    @Override
    public byte[] signRsa(byte[] message, PrivateKey prk) {
        return performOperationSafe(() -> {
            Signature signer = Signature.getInstance("SHA256withRSAAndMGF1", provider);
            signer.initSign(prk);
            signer.update(message);
            return signer.sign();
        }, "Error during RSA signing");
    }

    @Override
    public byte[] signEcdsa(byte[] message, PrivateKey prk) {
        return performOperationSafe(() -> {
            Signature signer = Signature.getInstance("SHA256withECDSA", provider);
            signer.initSign(prk);
            signer.update(message);
            return signer.sign();
        }, "Error during ECDSA signing");
    }

    @Override
    public byte[] deriveKeyAndDecrypt(PrivateKey selfPrivateKey, PublicKey otherPublicKey, byte[] message) {
        return performOperationSafe(() -> {
            byte[] aesKey = deriveAesEncryptionKey(selfPrivateKey, otherPublicKey);
            return decryptUsingAesGcm(message, aesKey, null);
        }, "Error during ECIES encryption");
    }

    private byte[] generateIV() {
        byte[] iv = new byte[GCM_IV_LENGTH];
        ThreadLocalRandom.current().nextBytes(iv);
        return iv;
    }

    private interface UnsafeSupplier<T> {
        T get() throws Exception;
    }

    private interface UnsafeOperation {
        void perform() throws Exception;
    }
}
