/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import com.fasterxml.jackson.annotation.JsonInclude;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.SerializationFeature;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.AesGcmEncryptionData;
import de.gematik.ti.sgd.domain.SgdInstance;
import de.gematik.ti.sgd.domain.configuration.SgdClientProperties;
import de.gematik.ti.sgd.domain.json.*;
import de.gematik.ti.sgd.domain.sgdCommunication.vectors.SgdVector;
import de.gematik.ti.sgd.domain.util.PHRUtils;
import de.gematik.ti.sgd.exceptions.client.InvalidStructureException;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import de.gematik.ws.fa.phr.v1.PHRKeyType;
import de.gematik.ws.fd.phrs.authorizationservice.v1.EncryptedKeyContainerType;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.security.KeyPair;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Optional;
import java.util.function.Supplier;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;

public class SgdClient {
    private static final Logger LOG = LoggerFactory.getLogger(SgdClient.class);

    private final X509Certificate clientCertificate;
    private final KeyPair clientKeyPair;
    private final List<PublicKey> possibleSigners;
    private final ObjectMapper objectMapper = buildObjectMapper();
    private AesGcmEncryptionData sgd1AesGcmKey = null;
    private AesGcmEncryptionData sgd2AesGcmKey = null;
    private Sgd1ServerWrapper server1Proxy;
    private Sgd2ServerWrapper server2Proxy;

    public SgdClient(SgdClientProperties sgdClientProperties, Sgd1ServerWrapper server1Proxy, Sgd2ServerWrapper server2Proxy) {
        this.clientCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromCrtFile(sgdClientProperties.getClientCertificateFile());
        try {
            this.clientKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(sgdClientProperties.getClientKeyPairFile(), "00", "00");
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
        this.possibleSigners = extractPublicKeysFromListOfCertificateFilenames(sgdClientProperties.getPossibleSignerCertificateFiles());
        this.server1Proxy = server1Proxy;
        this.server2Proxy = server2Proxy;
    }

    public SgdClient(X509Certificate clientCertificate, KeyPair clientKeyPair, List<String> possibleSignerCertificateFileNames, Sgd1ServerWrapper server1Proxy, Sgd2ServerWrapper server2Proxy) {
        this.clientCertificate = clientCertificate;
        this.clientKeyPair = clientKeyPair;
        this.possibleSigners = extractPublicKeysFromListOfCertificateFilenames(possibleSignerCertificateFileNames);
        this.server1Proxy = server1Proxy;
        this.server2Proxy = server2Proxy;
    }

    private ObjectMapper buildObjectMapper() {
        ObjectMapper objectMapper = new ObjectMapper();
        objectMapper.enable(SerializationFeature.INDENT_OUTPUT);
        objectMapper.setSerializationInclusion(JsonInclude.Include.NON_NULL);
        return objectMapper;
    }

    public void resetSgdResults() {
        sgd1AesGcmKey = null;
        sgd2AesGcmKey = null;
    }

    public AesGcmEncryptionData getSgd1AesGcmKey() {
        return sgd1AesGcmKey;
    }

    public AesGcmEncryptionData getSgd2AesGcmKey() {
        return sgd2AesGcmKey;
    }

    public void performInitialSgdKeyDerivation() {
        performKeyDerivationForVector(Optional.empty());
    }

    public SgdClient performKeyDerivationForVector(final Optional<SgdVector> vectorOptional) {
        return retryTargetCallIfTransientServerException(() -> performKeyDerivationForVectorUnsafe(vectorOptional));
    }

    public <T> T retryTargetCallIfTransientServerException(final Supplier<T> targetCall) {
        int retryCounter = 5;
        SgdServerException lastServerException = new SgdServerException("Can not happen", SgdResponseStatus.MISSING_DATA);
        while (retryCounter > 0) {
            try {
                return targetCall.get();
            } catch (SgdServerException e) {
                if (e.getResponseStatus() == SgdResponseStatus.RESTART_PROTOCOL
                        || e.getResponseStatus() == SgdResponseStatus.OCSP_RESPONSE_NOT_AVAILABLE) {
                    lastServerException = e;
                    retryCounter--;
                } else {
                    throw e;
                }
            }
        }
        throw lastServerException;
    }

    private SgdClient performKeyDerivationForVectorUnsafe(final Optional<SgdVector> vectorOptional) {
        SgdClientKeyDerivationMachine stateMachine = createNewStateMachineAndExchangePublicKeysAndAuthenticationToken();

        vectorOptional.ifPresent(vector -> {
            stateMachine.setSgd1AssociatedData(vector.toString().getBytes());
            stateMachine.setSgd2AssociatedData(vector.toString().getBytes());
        });

        stateMachine.receiveKeyDerivationResponse_sgd1(
                performEncryptedMessageExchange(
                        stateMachine.getKeyDerivationRequest_Sgd1(), SgdInstance.SGD_1));
        stateMachine.receiveKeyDerivationResponse_sgd2(
                performEncryptedMessageExchange(
                        stateMachine.getKeyDerivationRequest_Sgd2(), SgdInstance.SGD_2));

        sgd1AesGcmKey = stateMachine.getSgd1AesGcmKey();
        sgd2AesGcmKey = stateMachine.getSgd2AesGcmKey();

        return this;
    }

    public PHRKeyType decipherEncryptedKeyContainer(final EncryptedKeyContainerType doubleEncryptedKeyContainer) {
        return retryTargetCallIfTransientServerException(() -> decipherEncryptedKeyContainerUnsafe(doubleEncryptedKeyContainer));
    }

    private PHRKeyType decipherEncryptedKeyContainerUnsafe(final EncryptedKeyContainerType doubleEncryptedKeyContainer) {
        SgdClientKeyDerivationMachine stateMachine = createNewStateMachineAndExchangePublicKeysAndAuthenticationToken();

        final List<byte[]> adList = convertAssociatedDataToByteArrayList(doubleEncryptedKeyContainer);
        final String associatedDataGcm = joinByteArrayListToString(adList);
        if (adList.size() != 2) {
            throw new InvalidStructureException("Ill-formatted encryptedKeyContainer encountered!");
        }

        stateMachine.setSgd2AssociatedData(adList.get(1));
        stateMachine.setSgd1AssociatedData(adList.get(0));

        stateMachine.receiveKeyDerivationResponse_sgd2(
                performEncryptedMessageExchange(
                        stateMachine.getKeyDerivationRequest_Sgd2(), SgdInstance.SGD_2));

        stateMachine.receiveKeyDerivationResponse_sgd1(
                performEncryptedMessageExchange(
                        stateMachine.getKeyDerivationRequest_Sgd1(), SgdInstance.SGD_1));

        return PHRUtils.decipherTwiceFromContainer(doubleEncryptedKeyContainer, stateMachine.getSgd1AesGcmKey().getAesKey(), stateMachine.getSgd2AesGcmKey().getAesKey(), associatedDataGcm.getBytes());
    }

    public String joinByteArrayListToString(List<byte[]> adList) {
        return adList.stream()
                .map(String::new)
                .collect(Collectors.joining());
    }

    public List<byte[]> convertAssociatedDataToByteArrayList(EncryptedKeyContainerType doubleEncryptedKeyContainer) {
        return Stream.of(doubleEncryptedKeyContainer.getAssociatedData().split("\\s+"))
                .map(String::trim)
                .map(b64EncodedStr -> SgdCryptoUtilities.getInstance().convertB64StringToBytes((b64EncodedStr)))
                .collect(Collectors.toList());
    }

    private SgdClientKeyDerivationMachine createNewStateMachineAndExchangePublicKeysAndAuthenticationToken() {
        SgdClientKeyDerivationMachine stateMachine = new SgdClientKeyDerivationMachine(clientCertificate, clientKeyPair, possibleSigners);

        stateMachine.receivePublicKeyResponse_sgd1(
                performPublicKeyExchange(
                        stateMachine.getPublicKeyRequest(), SgdInstance.SGD_1));
        stateMachine.receivePublicKeyResponse_sgd2(
                performPublicKeyExchange(
                        stateMachine.getPublicKeyRequest(), SgdInstance.SGD_2));

        stateMachine.receiveAuthenticationTokenResponse_sgd1(
                performEncryptedMessageExchange(
                        stateMachine.getAuthenticationTokenRequest_sgd1(), SgdInstance.SGD_1));
        stateMachine.receiveAuthenticationTokenResponse_sgd2(
                performEncryptedMessageExchange(
                        stateMachine.getAuthenticationTokenRequest_sgd2(), SgdInstance.SGD_2));

        return stateMachine;
    }

    private EncryptedMessageContainer performEncryptedMessageExchange(SgdRequest sgdRequest, SgdInstance instance) {
        final EncryptedMessageContainer encryptedMessageContainer = performSgdExchange(sgdRequest, instance).toEncryptedMessageContainer();
        return encryptedMessageContainer;
    }

    private GetPublicKeyResponse performPublicKeyExchange(SgdRequest sgdRequest, SgdInstance instance) {
        final GetPublicKeyResponse getPublicKeyResponse = performSgdExchange(sgdRequest, instance).toGetPublicKeyResponse();
        return getPublicKeyResponse;
    }

    private SgdResponse performSgdExchange(SgdRequest sgdRequest, SgdInstance instance) {
        doLoggingForOutgoingRequest(sgdRequest, instance);
        SgdResponse serverReply = retrieveResultFromSgd(sgdRequest, instance);

        doLoggingForIncomingReplies(serverReply, instance);
        if (serverReply.getStatus() != null
                && serverReply.getStatus() != SgdResponseStatus.OK) {
            convertToAndThrowSgdServerException(serverReply);
        }
        return serverReply;
    }

    private void convertToAndThrowSgdServerException(SgdResponse serverReply) {
        throw new SgdServerException(serverReply);
    }

    private void doLoggingForIncomingReplies(Object result, SgdInstance instance) {
        try {
            LOG.info("Got reply from " + instance + ": " + objectMapper.writeValueAsString(result));
        } catch (JsonProcessingException e) {
            // swallow
        }
    }

    private void doLoggingForOutgoingRequest(SgdRequest sgdRequest, SgdInstance instance) {
        try {
            LOG.info("Sending Request to " + instance + ": " + objectMapper.writeValueAsString(sgdRequest));
        } catch (JsonProcessingException e) {
            // swallow
        }
    }

    private SgdResponse retrieveResultFromSgd(SgdRequest sgdRequest, SgdInstance instance) {
        if (instance == SgdInstance.SGD_1) {
            return server1Proxy.sgd1Endpoint(sgdRequest);
        } else {
            return server2Proxy.sgd2Endpoint(sgdRequest);
        }
    }
}
