/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.AesGcmEncryptionData;
import de.gematik.ti.sgd.domain.SgdInstance;
import de.gematik.ti.sgd.domain.json.EncryptedMessageContainer;
import de.gematik.ti.sgd.domain.json.GetPublicKeyResponse;
import de.gematik.ti.sgd.domain.json.SgdCommand;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.domain.sgdCommunication.EncryptedMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.KeyDerivationResponseSgdMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import de.gematik.ti.sgd.domain.sgdCommunication.SgdMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.vectors.SgdVector;
import org.apache.commons.codec.digest.DigestUtils;
import org.apache.commons.lang3.ArrayUtils;
import org.apache.commons.lang3.RandomStringUtils;

import javax.xml.bind.DatatypeConverter;
import java.security.KeyPair;
import java.security.PublicKey;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;

public class SgdClientKeyDerivationMachine {
    // Initale Daten
    private final X509Certificate clientCertificate;
    private final KeyPair clientCertificateKeyPair;
    private final List<PublicKey> possibleSigners;
    private Optional<String> associatedDataOptionalSgd1 = Optional.empty();
    private Optional<String> associatedDataOptionalSgd2 = Optional.empty();

    // GetPublicKey
    private PublicKeyEcies publicKeyEcies_sgd1;
    private PublicKeyEcies publicKeyEcies_sgd2;

    // Zufallsdaten pro Anfrage
    private byte[] challenge = new byte[32];
    private KeyPair ephemeralEciesKeyPair;
    private byte[] sgd1AuthenticationToken;
    private byte[] sgd2AuthenticationToken;
    private String requestId = RandomStringUtils.random((256 / 8) * 2, '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', 'a', 'b', 'c', 'd', 'e', 'f');

    // Ergebnisse
    private AesGcmEncryptionData sgd1AesGcmKey;
    private AesGcmEncryptionData sgd2AesGcmKey;

    public SgdClientKeyDerivationMachine(X509Certificate clientCertificate, KeyPair clientCertificateKeyPair, List<PublicKey> possibleSigners) {
        this.clientCertificate = clientCertificate;
        this.clientCertificateKeyPair = clientCertificateKeyPair;
        this.possibleSigners = possibleSigners;
        ThreadLocalRandom.current().nextBytes(challenge);
        try {
            ephemeralEciesKeyPair = SgdCryptoUtilities.getInstance().generateEccKeyPair();
        } catch (Exception e) {
            throw new RuntimeException("Error while instantiating SgdClientKeyDerivationMachine: Could not generate Ephemeral ECC Key Pair", e);
        }
    }

    public SgdRequest getPublicKeyRequest() {
        return SgdClientUtils.generateGetPublicKeyRequest(clientCertificate);
    }

    public void receivePublicKeyResponse_sgd1(GetPublicKeyResponse getPublicKeyResponse) {
        SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(getPublicKeyResponse, possibleSigners, SgdInstance.SGD_1);

        publicKeyEcies_sgd1 = getPublicKeyResponse.getPublicKeyEcies();
    }

    public void receivePublicKeyResponse_sgd2(GetPublicKeyResponse getPublicKeyResponse) {
        SgdClientUtils.performGetPublicKeyVerificationAndThrowExceptionWhenFail(getPublicKeyResponse, possibleSigners, SgdInstance.SGD_2);

        publicKeyEcies_sgd2 = getPublicKeyResponse.getPublicKeyEcies();
    }

    public SgdRequest getAuthenticationTokenRequest_sgd1() {
        final String payload = "Challenge " + SgdCryptoUtilities.getInstance().convertToLowerCaseHexWith64ByteLength(challenge);
        return generateVerifiedAndEncryptedSgdRequest(publicKeyEcies_sgd1.getPublicKey(), payload, SgdCommand.GET_AUTHENTICATION_TOKEN);
    }

    public SgdRequest getAuthenticationTokenRequest_sgd2() {
        final String payload = "Challenge " + SgdCryptoUtilities.getInstance().convertToLowerCaseHexWith64ByteLength(challenge);
        return generateVerifiedAndEncryptedSgdRequest(publicKeyEcies_sgd2.getPublicKey(), payload, SgdCommand.GET_AUTHENTICATION_TOKEN);
    }

    public void receiveAuthenticationTokenResponse_sgd1(EncryptedMessageContainer authenticationToken) {
        sgd1AuthenticationToken = checkAuthenticationTokenResponseAndReturnToken(authenticationToken);
    }

    public void receiveAuthenticationTokenResponse_sgd2(EncryptedMessageContainer authenticationToken) {
        sgd2AuthenticationToken = checkAuthenticationTokenResponseAndReturnToken(authenticationToken);
    }

    public byte[] checkAuthenticationTokenResponseAndReturnToken(EncryptedMessageContainer authenticationToken) {
        String payload = new String(authenticationToken.getEncryptedMessage().decryptMessage(ephemeralEciesKeyPair));
        String[] payloadParts = payload.split(" ");

        performA18028ChecksAndThrowExceptionInCaseOfFailure(payloadParts);

        return DatatypeConverter.parseHexBinary(payloadParts[3].substring("AT".length()));
    }

    private void performA18028ChecksAndThrowExceptionInCaseOfFailure(String[] payloadParts) {
        // Check aus A_18028
        checkArgument(payloadParts.length == 4, "Die GetAuthenticationToken-Antwort muss aus genau vier Teilen bestehen!");
        checkArgument(payloadParts[0].equals("Response"), "Die GetAuthenticationToken-Antwort muss mit 'Response' beginnen!");
        if (!Arrays.equals(DatatypeConverter.parseHexBinary(payloadParts[1]), challenge)) {
            throw new IllegalArgumentException("Die Client-Challenge ist nicht die originäre! Erwartet wurde '" + DatatypeConverter.printHexBinary(challenge) + "', gefunden wurde '" + payloadParts[1] + "'!");
        }
        try {
            final byte[] h_computedValue = DigestUtils.sha256(ArrayUtils.addAll(new PublicKeyEcies(ephemeralEciesKeyPair.getPublic()).getEciesPublicKeyAsString().getBytes(), this.clientCertificate.getEncoded()));
            if (!Arrays.equals(h_computedValue, DatatypeConverter.parseHexBinary(payloadParts[2]))) {
                throw new IllegalArgumentException("Der h-Value stimmt nicht mit der Erwartung überein! Erwartet wurde '" + DatatypeConverter.printHexBinary(h_computedValue) + "', gefunden wurde '" + payloadParts[2] + "'");
            }
        } catch (CertificateEncodingException e) {
            throw new RuntimeException(e);
        }
        checkArgument(payloadParts[3].startsWith("AT"), "Der Authentication-Token muss mit 'AT' beginnen. Der Token ist aber wie folgt '" + payloadParts[3] + "'!");
        checkArgument(payloadParts[3].substring(2).matches("[0-9a-f]{64}"), "Der Authentication-Token '" + payloadParts[3] + "' ist kein lower-case 256-Bit Hex-String!");
    }

    private void checkArgument(boolean condition, String message) {
        if (!condition) {
            throw new RuntimeException(message);
        }
    }

    public SgdRequest getKeyDerivationRequest_Sgd1() {
        return getKeyDerivationRequest_Sgd1(
                associatedDataOptionalSgd1
                        .map(associatedData -> new SgdVector(associatedData))
                        .orElse(SgdVector.createForR1InitialDerivation(getKvnr())));
    }

    public SgdRequest getKeyDerivationRequest_Sgd1(SgdVector sgdVector) {
        return constructKeyDerivationRequest(sgdVector, publicKeyEcies_sgd1.getPublicKey(), sgd1AuthenticationToken);
    }

    public SgdRequest getKeyDerivationRequest_Sgd2() {
        return getKeyDerivationRequest_Sgd2(
                associatedDataOptionalSgd2
                        .map(associatedData -> new SgdVector(associatedData))
                        .orElse(SgdVector.createForR1InitialDerivation(getKvnr())));
    }

    public SgdRequest getKeyDerivationRequest_Sgd2(SgdVector sgdVector) {
        return constructKeyDerivationRequest(sgdVector, publicKeyEcies_sgd2.getPublicKey(), sgd2AuthenticationToken);
    }

    public SgdRequest constructKeyDerivationRequest(SgdVector sgdVector, PublicKey publicKey, byte[] sgd1AuthenticationToken) {
        return generateVerifiedAndEncryptedSgdRequest(publicKey,
                "AT" + SgdCryptoUtilities.getInstance().convertToLowerCaseHexWith64ByteLength(sgd1AuthenticationToken) + " "
                        + requestId + " "
                        + "KeyDerivation "
                        + sgdVector.toString(), SgdCommand.KEY_DERIVATION);
    }

    private String getKvnr() {
        return SgdCryptoUtilities.extractKVNRfromSubjectDN(this.clientCertificate.getSubjectDN().getName());
    }

    private SgdRequest generateVerifiedAndEncryptedSgdRequest(PublicKey sgdEphemeralPublicKey, String payload, SgdCommand command) {
        try {
            if (publicKeyEcies_sgd1 == null) {
                throw new RuntimeException("Could not generate Encrypted SGD Request: No public Key for SGD1 retrieved yet!");
            }
            if (publicKeyEcies_sgd2 == null) {
                throw new RuntimeException("Could not generate Encrypted SGD Request: No public Key for SGD2 retrieved yet!");
            }
            final PublicKeyEcies publicKeyEcies = new PublicKeyEcies(ephemeralEciesKeyPair.getPublic(), publicKeyEcies_sgd1.getPublicKey(), publicKeyEcies_sgd2.getPublicKey());
            return SgdRequest.builder()
                    .command(command)
                    .certificate(DatatypeConverter.printBase64Binary(clientCertificate.getEncoded()))
                    .publicKeyEcies(publicKeyEcies)
                    .signature(publicKeyEcies.getSignatureAsBase64EncodedString(clientCertificateKeyPair.getPrivate()))
                    .encryptedMessage(new EncryptedMessage(sgdEphemeralPublicKey, payload))
                    .build();
        } catch (CertificateEncodingException e) {
            throw new RuntimeException("Error while encoding GetAuthenticationTokenRequest", e);
        }
    }

    public void receiveKeyDerivationResponse_sgd1(EncryptedMessageContainer keyDerivationResponse) {
        sgd1AesGcmKey = convertAndCheckKeyDerivationResponse(keyDerivationResponse);
    }

    public void receiveKeyDerivationResponse_sgd2(EncryptedMessageContainer keyDerivationResponse) {
        sgd2AesGcmKey = convertAndCheckKeyDerivationResponse(keyDerivationResponse);
    }

    private AesGcmEncryptionData convertAndCheckKeyDerivationResponse(EncryptedMessageContainer keyDerivationResponse) {
        final SgdMessage sgdMessage = keyDerivationResponse.getEncryptedMessage().decryptToSgdMessage(ephemeralEciesKeyPair);
        if (!(sgdMessage instanceof KeyDerivationResponseSgdMessage)) {
            throw new RuntimeException("Unexpected SgdMessage: Expected KeyDerivationResponse");
        }
        KeyDerivationResponseSgdMessage message = (KeyDerivationResponseSgdMessage) sgdMessage;
        if (!message.getRequestId().equalsIgnoreCase(requestId)) {
            throw new RuntimeException("RequestId does not match for KeyDerivationResponse");
        }
        return AesGcmEncryptionData.AesGcmEncryptionDataBuilder.anAesGcmEncryptionData()
                .withAesKey(DatatypeConverter.parseHexBinary(message.getAesKeyAsHexString()))
                .withAssociatedData(message.getSgdVector().toString())
                .build();
    }

    public AesGcmEncryptionData getSgd1AesGcmKey() {
        return sgd1AesGcmKey;
    }

    public AesGcmEncryptionData getSgd2AesGcmKey() {
        return sgd2AesGcmKey;
    }

    public void setSgd1AssociatedData(byte[] associatedData) {
        this.associatedDataOptionalSgd1 = Optional.of(new String(associatedData));
    }

    public void setSgd2AssociatedData(byte[] associatedData) {
        this.associatedDataOptionalSgd2 = Optional.of(new String(associatedData));
    }
}
