/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.exceptions.generic.CertificateValidationException;
import de.gematik.ti.sgd.exceptions.server.ServerInvalidCertificateException;
import org.apache.commons.lang3.StringUtils;

import java.security.Principal;
import java.security.PublicKey;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Objects;
import java.util.Optional;

import static de.gematik.ti.sgd.components.CertificateVerificationUtils.*;

public class SgdServerUtils {

    // gemSpec_SGD_ePA, A_17919
    public static void performGetPublicKeyVerificationAndThrowExceptionWhenFail(SgdRequest request, List<PublicKey> possibleSigners, List<X509Certificate> possibleOcspSignerCertificates) {
        X509Certificate clientAutCertificate;
        try {
            clientAutCertificate = SgdCryptoUtilities.getInstance().parseX509CertificateFromB64String(request.getCertificate());
        } catch (RuntimeException e) {
            throw new ServerInvalidCertificateException("Error while validating Client Certificate for getPublicKeyRequest", e);
        }

        try {
            assertThatCertificateIsValid(clientAutCertificate);
            assertThatCertificateSignatureIsValid(clientAutCertificate, possibleSigners);
            assertThatCertificateHasKvnrOrTelematikId(clientAutCertificate);
        } catch (CertificateValidationException e) {
            throw new ServerInvalidCertificateException("Error while validating Client Certificate for getPublicKeyRequest", e);
        }

        verifyOcspResponseIfPresent(request, possibleOcspSignerCertificates);
    }

    private static void verifyOcspResponseIfPresent(SgdRequest request, List<X509Certificate> possibleOcspSignerCertificates) {
        if (StringUtils.isEmpty(request.getOcspResponse())) {
            return;
        }

        try {
            byte[] ocspResponsePayload = SgdCryptoUtilities.getInstance().convertB64StringToBytes(request.getOcspResponse());
            SgdCryptoUtilities.getInstance().verifyOcspSignatureAndThrowExceptionWhenFail(ocspResponsePayload, possibleOcspSignerCertificates);
        } catch (Exception e) {
            throw new ServerInvalidCertificateException("Error while validating OCSP Response for getPublicKeyRequest", e);
        }
    }

    public static String perform18003Conversion(String value) {
        if (!value.contains(":")) {
            return value;
        }
        return "*" + SgdCryptoUtilities.getInstance().convertByteArrayToHexString(value.getBytes()).toLowerCase();
    }

    public static Optional<String> extractKVNRfromCertificate(X509Certificate certificate) {
        try {
            return Optional.ofNullable(certificate.getSubjectDN())
                    .map(Principal::getName)
                    .map(SgdCryptoUtilities::extractKVNRfromSubjectDN);
        } catch (Exception e) {
            return Optional.empty();
        }
    }
}
