/*
 * Copyright (c) 2019 gematik - Gesellschaft für Telematikanwendungen der Gesundheitskarte mbH
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package de.gematik.ti.sgd.components;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.jayway.jsonpath.DocumentContext;
import com.jayway.jsonpath.JsonPath;
import de.gematik.ti.sgd.SgdTestCryptoUtilities;
import de.gematik.ti.sgd.crypto.SgdCryptoUtilities;
import de.gematik.ti.sgd.domain.AesGcmEncryptionData;
import de.gematik.ti.sgd.domain.RecordAndContextKey;
import de.gematik.ti.sgd.domain.json.EncryptedMessageContainer;
import de.gematik.ti.sgd.domain.json.SgdRequest;
import de.gematik.ti.sgd.domain.json.SgdResponseStatus;
import de.gematik.ti.sgd.domain.sgdCommunication.KeyDerivationResponseSgdMessage;
import de.gematik.ti.sgd.domain.sgdCommunication.PublicKeyEcies;
import de.gematik.ti.sgd.domain.sgdCommunication.vectors.SgdVector;
import de.gematik.ti.sgd.domain.util.PHRUtils;
import de.gematik.ti.sgd.exceptions.server.SgdServerException;
import de.gematik.ws.fa.phr.v1.PHRKeyType;
import de.gematik.ws.fd.phrs.authorizationservice.v1.EncryptedKeyContainerType;
import org.apache.commons.codec.digest.DigestUtils;
import org.junit.Before;
import org.junit.Test;
import org.springframework.test.util.ReflectionTestUtils;

import javax.xml.bind.DatatypeConverter;
import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.security.KeyPair;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPublicKey;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ThreadLocalRandom;
import java.util.stream.Collectors;

import static de.gematik.ti.sgd.components.SgdClientUtils.extractPublicKeysFromListOfCertificateFilenames;
import static de.gematik.ti.sgd.crypto.SgdCryptoUtilities.extractKVNRfromSubjectDN;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

public class TestSgdCommunication {
    private static final String RSA_CLIENT_P12_FILE = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut.p12";
    private static List<String> SIGNER_LIST;

    static {
        SgdCryptoUtilities.setInstance(new SgdTestCryptoUtilities());
    }

    private final String P12_1_FILENAME = "src/test/resources/keys/egk/109500969_X114428530_c.ch.enc-ecc.p12";
    private final String P12_2_FILENAME = "src/test/resources/keys/egk/109500969_X114428530_c.ch.aut-ecc.p12";
    private final String P12_3_FILENAME = "src/test/resources/keys/egk/109500969_X114428531_c.ch.aut-ecc.p12";
    private final String P12_4_FILENAME = "src/test/resources/keys/egk/109500969_X114428533_c.ch.aut-ecc.p12";
    private final String P12_SMCB_FILENAME = "src/test/resources/keys/smcb/833621999741621_c.hci.aut-arge-psych-ecc.p12";

    private SgdClientKeyDerivationMachine client;
    private SchluesselGenerierungsDienst sgd1;
    private SchluesselGenerierungsDienst sgd2;

    private X509Certificate sgd1_Certificate;
    private X509Certificate sgd2_Certificate;
    private KeyPair sgd1_Key;
    private KeyPair sgd2_Key;
    private KeyPair sgd1EphemeralKeyPair;
    private KeyPair sgd2EphemeralKeyPair;
    private KeyPair clientEphemeralKeyPair;
    private KeyPair clientCertificateKeyPair;
    private X509Certificate clientCertificate;
    private KeyPair smcbKeyPair;
    private X509Certificate smcbCertificate;
    private byte[] clientChallenge;

    @Before
    public void setup() throws IOException {
        SIGNER_LIST = Files.list(Path.of("src/test/resources/keys/ca/"))
                .map(path -> path.toString())
                .collect(Collectors.toList());

        sgd1_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00");
        sgd2_Certificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00");
        sgd1_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd1.hsm-aut-ecc.p12", "00", "00");
        sgd2_Key = SgdCryptoUtilities.getInstance().loadKeypairFromP12File("src/test/resources/keys/sgd/c.sgd2.hsm-aut-ecc.p12", "00", "00");
        sgd1EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_1_FILENAME, "00", "00");
        sgd2EphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_4_FILENAME, "00", "00");
        clientEphemeralKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_2_FILENAME, "00", "00");
        clientCertificateKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_3_FILENAME, "00", "00");
        clientCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File(P12_3_FILENAME, "00");
        smcbKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(P12_SMCB_FILENAME, "00", "00");
        smcbCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File(P12_SMCB_FILENAME, "00");

        client = new SgdClientKeyDerivationMachine(clientCertificate, clientCertificateKeyPair, extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));
        ReflectionTestUtils.setField(client, "ephemeralEciesKeyPair", clientEphemeralKeyPair);
        clientChallenge = (byte[]) ReflectionTestUtils.getField(client, "challenge");

        sgd1 = new SchluesselGenerierungsDienst(sgd1_Certificate, sgd1_Key.getPrivate(), sgd1EphemeralKeyPair);
        sgd2 = new SchluesselGenerierungsDienst(sgd2_Certificate, sgd2_Key.getPrivate(), sgd2EphemeralKeyPair);
        ReflectionTestUtils.setField(sgd1, "possibleCertificateSigners", extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));
        ReflectionTestUtils.setField(sgd2, "possibleCertificateSigners", extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));
    }

    @Test
    public void happyPath() {
        // getPublicKey
        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        // getAuthenticationToken
        // Request
        final SgdRequest authenticationTokenRequestSgd1 = client.getAuthenticationTokenRequest_sgd1();
        final SgdRequest authenticationTokenRequestSgd2 = client.getAuthenticationTokenRequest_sgd2();
        checkAuthenticationTokenRequestAndSerialization(authenticationTokenRequestSgd1);
        checkAuthenticationTokenRequestAndSerialization(authenticationTokenRequestSgd2);
        // Response
        final EncryptedMessageContainer sgd1AuthenticationToken = sgd1.receive(authenticationTokenRequestSgd1).toEncryptedMessageContainer();
        final EncryptedMessageContainer sgd2AuthenticationToken = sgd2.receive(authenticationTokenRequestSgd2).toEncryptedMessageContainer();
        checkGetAuthenticationTokenResponse(sgd1AuthenticationToken);
        checkGetAuthenticationTokenResponse(sgd2AuthenticationToken);
        client.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken);
        client.receiveAuthenticationTokenResponse_sgd2(sgd2AuthenticationToken);

        // keyDerivation (r1, initial)
        final SgdRequest keyDerivationRequestSgd1 = client.getKeyDerivationRequest_Sgd1();
        final SgdRequest keyDerivationRequestSgd2 = client.getKeyDerivationRequest_Sgd2();
        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(keyDerivationRequestSgd1).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(keyDerivationRequestSgd2).toEncryptedMessageContainer();
        checkKeyDerivationResponse(keyDerivationResponseSgd1, clientCertificate);
        checkKeyDerivationResponse(keyDerivationResponseSgd2, clientCertificate);
        client.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        client.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);
    }

    @Test
    public void sgd1And2AreExchanged_clientShouldGiveException() {
        assertThatThrownBy(() -> client.receivePublicKeyResponse_sgd1(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse()))
                .isInstanceOf(RuntimeException.class);
    }

    @Test
    public void deriveKeysWithGivenAssociatedData_ResultsShouldMatch() {
        final AesGcmEncryptionData sgd1InitialResult = sgd1.offlineKeyDerivationForKvnr(getClientKvnr(clientCertificate));
        final AesGcmEncryptionData sgd2InitialResult = sgd2.offlineKeyDerivationForKvnr(getClientKvnr(clientCertificate));

        ReflectionTestUtils.setField(client, "associatedDataOptionalSgd1", Optional.of(sgd1InitialResult.getAssociatedData()));
        ReflectionTestUtils.setField(client, "associatedDataOptionalSgd2", Optional.of(sgd2InitialResult.getAssociatedData()));

        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        client.receiveKeyDerivationResponse_sgd1(sgd1.receive(client.getKeyDerivationRequest_Sgd1()).toEncryptedMessageContainer());
        client.receiveKeyDerivationResponse_sgd2(sgd2.receive(client.getKeyDerivationRequest_Sgd2()).toEncryptedMessageContainer());

        assertThat(client.getSgd1AesGcmKey().getAesKey())
                .containsExactly(sgd1InitialResult.getAesKey());
        assertThat(client.getSgd2AesGcmKey().getAesKey())
                .containsExactly(sgd2InitialResult.getAesKey());

        assertThat(client.getSgd1AesGcmKey().getAssociatedData())
                .isEqualTo(sgd1InitialResult.getAssociatedData());
        assertThat(client.getSgd2AesGcmKey().getAssociatedData())
                .isEqualTo(sgd2InitialResult.getAssociatedData());
    }

    @Test
    public void r2Initial() {
        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        final String otherKvnr = "A1234567890";
        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(client.getKeyDerivationRequest_Sgd1(SgdVector.createForR2InitialDerivation(otherKvnr))).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(client.getKeyDerivationRequest_Sgd2(SgdVector.createForR2InitialDerivation(otherKvnr))).toEncryptedMessageContainer();

        assertThat(keyDerivationResponseSgd1.getDebug_encryptedMessageCleartext())
                .contains(" r2:");

        client.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        client.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);

        assertThat(client.getSgd1AesGcmKey().getAssociatedData())
                .startsWith("r2:")
                .contains(getClientKvnr(clientCertificate) + ":" + otherKvnr);
        assertThat(client.getSgd2AesGcmKey().getAssociatedData())
                .startsWith("r2:")
                .contains(getClientKvnr(clientCertificate) + ":" + otherKvnr);
    }

    @Test
    public void r2Repeated() {
        final String otherKvnr = "A1234567890";
        final String clientKvnr = getClientKvnr(clientCertificate);
        final AesGcmEncryptionData sgd1InitialResult = sgd1.offlineKeyDerivationForKvnr(SgdVector.createForR2RepeatedDerivation(otherKvnr, clientKvnr, sgd1.getCurrentDerivationKeyName()));
        final AesGcmEncryptionData sgd2InitialResult = sgd2.offlineKeyDerivationForKvnr(SgdVector.createForR2RepeatedDerivation(otherKvnr, clientKvnr, sgd2.getCurrentDerivationKeyName()));

        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(client.getKeyDerivationRequest_Sgd1(new SgdVector(sgd1InitialResult.getAssociatedData()))).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(client.getKeyDerivationRequest_Sgd2(new SgdVector(sgd2InitialResult.getAssociatedData()))).toEncryptedMessageContainer();

        assertThat(keyDerivationResponseSgd1.getDebug_encryptedMessageCleartext())
                .contains(" r2:");

        client.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        client.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);

        assertThat(client.getSgd1AesGcmKey().getAesKey())
                .containsExactly(sgd1InitialResult.getAesKey());
        assertThat(client.getSgd2AesGcmKey().getAesKey())
                .containsExactly(sgd2InitialResult.getAesKey());
    }

    @Test
    public void r3Initial() {
        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        final String accountKvnr = "A1234567890";
        final String authorizedKvnr = "A0987654321";
        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(client.getKeyDerivationRequest_Sgd1(SgdVector.createForR3InitialDerivation(authorizedKvnr, accountKvnr))).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(client.getKeyDerivationRequest_Sgd2(SgdVector.createForR3InitialDerivation(authorizedKvnr, accountKvnr))).toEncryptedMessageContainer();

        assertThat(keyDerivationResponseSgd1.getDebug_encryptedMessageCleartext())
                .contains(" r3:");

        client.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        client.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);

        assertThat(client.getSgd1AesGcmKey().getAssociatedData())
                .startsWith("r3:")
                .contains(accountKvnr + ":" + getClientKvnr(clientCertificate) + ":" + authorizedKvnr);
        assertThat(client.getSgd2AesGcmKey().getAssociatedData())
                .startsWith("r3:")
                .contains(accountKvnr + ":" + getClientKvnr(clientCertificate) + ":" + authorizedKvnr);
    }

    @Test
    public void r3Repeated() {
        final String accountKvnr = "A1234567890";
        final String authorizedTelematikId = "1-2-EPA-833621999741621";
        final String clientKvnr = getClientKvnr(clientCertificate);
        final AesGcmEncryptionData sgd1InitialResult = sgd1.offlineKeyDerivationForKvnr(SgdVector.createForR3RepeatedDerivation(accountKvnr, clientKvnr, authorizedTelematikId, sgd1.getCurrentDerivationKeyName()));
        final AesGcmEncryptionData sgd2InitialResult = sgd2.offlineKeyDerivationForKvnr(SgdVector.createForR3RepeatedDerivation(accountKvnr, clientKvnr, authorizedTelematikId, sgd2.getCurrentDerivationKeyName()));
        client = new SgdClientKeyDerivationMachine(smcbCertificate, smcbKeyPair, extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));

        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(client.getKeyDerivationRequest_Sgd1(new SgdVector(sgd1InitialResult.getAssociatedData()))).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(client.getKeyDerivationRequest_Sgd2(new SgdVector(sgd2InitialResult.getAssociatedData()))).toEncryptedMessageContainer();

        assertThat(keyDerivationResponseSgd1.getDebug_encryptedMessageCleartext())
                .contains(" r3:");

        client.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        client.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);

        assertThat(client.getSgd1AesGcmKey().getAesKey())
                .containsExactly(sgd1InitialResult.getAesKey());
        assertThat(client.getSgd2AesGcmKey().getAesKey())
                .containsExactly(sgd2InitialResult.getAesKey());
    }

    @Test
    public void deciperEncryptedContainer() {
        final AesGcmEncryptionData sgd1InitialResult = sgd1.offlineKeyDerivationForKvnr(getClientKvnr(clientCertificate));
        final AesGcmEncryptionData sgd2InitialResult = sgd2.offlineKeyDerivationForKvnr(getClientKvnr(clientCertificate));
        RecordAndContextKey key = RecordAndContextKey.newRandomKeysForKvnr(getClientKvnr(clientCertificate));
        final EncryptedKeyContainerType encryptedKeyContainer = key.encipherTwiceToContainer(sgd1InitialResult, sgd2InitialResult);

        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        final String[] associatedData = encryptedKeyContainer.getAssociatedData().split("\\s+");
        client.setSgd2AssociatedData(DatatypeConverter.parseBase64Binary(associatedData[1]));
        client.setSgd1AssociatedData(DatatypeConverter.parseBase64Binary(associatedData[0]));
        client.receiveKeyDerivationResponse_sgd2(sgd2.receive(client.getKeyDerivationRequest_Sgd2()).toEncryptedMessageContainer());
        client.receiveKeyDerivationResponse_sgd1(sgd1.receive(client.getKeyDerivationRequest_Sgd1()).toEncryptedMessageContainer());

        final PHRKeyType recovered = PHRUtils.decipherTwiceFromContainer(
                encryptedKeyContainer,
                client.getSgd1AesGcmKey().getAesKey(),
                client.getSgd2AesGcmKey().getAesKey(),
                PHRUtils.buildOuterAesGcmAdFromOuterAssociatedData(encryptedKeyContainer.getAssociatedData()));

        assertThat(recovered.getContextKey().getValue()).containsExactly(key.getContextKeyBytes());
        assertThat(recovered.getRecordKey().getValue()).containsExactly(key.getRecordKeyBytes());
    }

    @Test
    public void happyPath_RSA() {
        KeyPair rsaClientCertificateKeyPair = SgdCryptoUtilities.getInstance().loadKeypairFromP12File(RSA_CLIENT_P12_FILE, "00", "00");
        X509Certificate rsaClientCertificate = SgdCryptoUtilities.getInstance().loadX509CertificateFromP12File(RSA_CLIENT_P12_FILE, "00");
        SgdClientKeyDerivationMachine rsaClient
                = new SgdClientKeyDerivationMachine(rsaClientCertificate, rsaClientCertificateKeyPair, extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));

        ReflectionTestUtils.setField(rsaClient, "ephemeralEciesKeyPair", clientEphemeralKeyPair);
        ReflectionTestUtils.setField(rsaClient, "challenge", clientChallenge);


        // getPublicKey
        rsaClient.receivePublicKeyResponse_sgd1(sgd1.receive(rsaClient.getPublicKeyRequest()).toGetPublicKeyResponse());
        rsaClient.receivePublicKeyResponse_sgd2(sgd2.receive(rsaClient.getPublicKeyRequest()).toGetPublicKeyResponse());


        // getAuthenticationToken
        // Request
        final SgdRequest authenticationTokenRequestSgd1 = rsaClient.getAuthenticationTokenRequest_sgd1();
        final SgdRequest authenticationTokenRequestSgd2 = rsaClient.getAuthenticationTokenRequest_sgd2();
        checkAuthenticationTokenRequestAndSerialization(authenticationTokenRequestSgd1);
        checkAuthenticationTokenRequestAndSerialization(authenticationTokenRequestSgd2);
        // Response
        final EncryptedMessageContainer sgd1AuthenticationToken = sgd1.receive(authenticationTokenRequestSgd1).toEncryptedMessageContainer();
        final EncryptedMessageContainer sgd2AuthenticationToken = sgd2.receive(authenticationTokenRequestSgd2).toEncryptedMessageContainer();
        checkGetAuthenticationTokenResponse(sgd1AuthenticationToken);
        checkGetAuthenticationTokenResponse(sgd2AuthenticationToken);
        rsaClient.receiveAuthenticationTokenResponse_sgd1(sgd1AuthenticationToken);
        rsaClient.receiveAuthenticationTokenResponse_sgd2(sgd2AuthenticationToken);


        // keyDerivation
        final SgdRequest keyDerivationRequestSgd1 = rsaClient.getKeyDerivationRequest_Sgd1();
        final SgdRequest keyDerivationRequestSgd2 = rsaClient.getKeyDerivationRequest_Sgd2();
        final EncryptedMessageContainer keyDerivationResponseSgd1 = sgd1.receive(keyDerivationRequestSgd1).toEncryptedMessageContainer();
        final EncryptedMessageContainer keyDerivationResponseSgd2 = sgd2.receive(keyDerivationRequestSgd2).toEncryptedMessageContainer();
        checkKeyDerivationResponse(keyDerivationResponseSgd1, rsaClientCertificate);
        checkKeyDerivationResponse(keyDerivationResponseSgd2, rsaClientCertificate);
        rsaClient.receiveKeyDerivationResponse_sgd1(keyDerivationResponseSgd1);
        rsaClient.receiveKeyDerivationResponse_sgd2(keyDerivationResponseSgd2);
    }

    @Test
    public void defunctCertificate_shouldGiveCertificateNotValid() {
        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        final SgdRequest authenticationTokenRequestSgd1 = client.getAuthenticationTokenRequest_sgd1();
        final String certificate = authenticationTokenRequestSgd1.getCertificate();
        authenticationTokenRequestSgd1.setCertificate(certificate.substring(0, certificate.length() - 5));
        assertThatThrownBy(() -> checkAuthenticationTokenRequestAndSerialization(authenticationTokenRequestSgd1))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.CERTIFICATE_NOT_VALID);
    }

    @Test
    public void smcbCertificate_shouldGiveCertificateNotValid() {
        client = new SgdClientKeyDerivationMachine(smcbCertificate, smcbKeyPair, extractPublicKeysFromListOfCertificateFilenames(SIGNER_LIST));

        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        assertThatThrownBy(() -> sgd1.receive(client.getKeyDerivationRequest_Sgd1()))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.KEY_DERIVATION_FAIL);
    }

    @Test
    public void alteredAuthenticationToken_ShouldGiveErrorForKeyDerivation() {
        client.receivePublicKeyResponse_sgd1(sgd1.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());
        client.receivePublicKeyResponse_sgd2(sgd2.receive(client.getPublicKeyRequest()).toGetPublicKeyResponse());

        client.receiveAuthenticationTokenResponse_sgd1(sgd1.receive(client.getAuthenticationTokenRequest_sgd1()).toEncryptedMessageContainer());
        client.receiveAuthenticationTokenResponse_sgd2(sgd2.receive(client.getAuthenticationTokenRequest_sgd2()).toEncryptedMessageContainer());

        byte[] result = new byte[512 / 8];
        ThreadLocalRandom.current().nextBytes(result);
        ReflectionTestUtils.setField(client, "sgd1AuthenticationToken", result);

        assertThatThrownBy(() -> sgd1.receive(client.getKeyDerivationRequest_Sgd1()))
                .isInstanceOf(SgdServerException.class)
                .hasFieldOrPropertyWithValue("responseStatus", SgdResponseStatus.KEY_DERIVATION_FAIL);
    }

    private void checkKeyDerivationResponse(EncryptedMessageContainer keyDerivationResponse, X509Certificate clientCertificate) {
        assertThat(keyDerivationResponse.getStatus())
                .isEqualTo(SgdResponseStatus.OK);
        String payload = new String(keyDerivationResponse.getEncryptedMessage().decryptMessage(clientEphemeralKeyPair));
        // A_18031
        assertThat(payload.split(" ").length)
                .isEqualTo(5);
        KeyDerivationResponseSgdMessage sgdMessage = (KeyDerivationResponseSgdMessage) keyDerivationResponse.getEncryptedMessage().decryptToSgdMessage(clientEphemeralKeyPair);
        assertThat(DatatypeConverter.parseHexBinary(sgdMessage.getAesKeyAsHexString()))
                .hasSize(256 / 8);
        final SgdVector sgdVector = sgdMessage.getSgdVector();
        assertThat(sgdVector.getKeyDerivationName())
                .isEqualTo(ReflectionTestUtils.getField(sgd1, "currentDerivationKeyName"));
        assertThat(sgdVector.getKvnr())
                .isEqualTo(getClientKvnr(clientCertificate));
    }

    private String getClientKvnr(X509Certificate clientCertificate) {
        return extractKVNRfromSubjectDN(clientCertificate.getSubjectDN().getName());
    }

    public void checkAuthenticationTokenRequestAndSerialization(SgdRequest request) {
        checkAuthenticationTokenRequestSerialization(request);
        testAuthenticationTokenRequestContent(request);
    }

    private void testAuthenticationTokenRequestContent(SgdRequest request) {
        final List<String> pubKeyParts = Arrays.asList(request.getPublicKeyEcies().getMessage().split(" "));
        assertThat(pubKeyParts.get(0))
                .isEqualToIgnoringCase("brainPoolP256r1");

        assertThat(pubKeyParts.get(1))
                .startsWith("0x");
        assertThat(pubKeyParts.get(2))
                .startsWith("0x");
        assertThat(request.getPublicKeyEcies().getPublicKey().getEncoded())
                .containsExactly(clientEphemeralKeyPair.getPublic().getEncoded());
        assertThat(pubKeyParts.get(3))
                .isEqualToIgnoringCase(DigestUtils.sha256Hex(new PublicKeyEcies(sgd1EphemeralKeyPair.getPublic()).getEciesPublicKeyAsString()));
        assertThat(pubKeyParts.get(4))
                .isEqualToIgnoringCase(DigestUtils.sha256Hex(new PublicKeyEcies(sgd2EphemeralKeyPair.getPublic()).getEciesPublicKeyAsString()));

    }

    private void checkAuthenticationTokenRequestSerialization(SgdRequest request) {
        final ObjectMapper mapper = new ObjectMapper();

        final DocumentContext result;
        try {
            result = JsonPath.parse(mapper.writeValueAsString(request));
        } catch (JsonProcessingException e) {
            throw new RuntimeException(e);
        }

        assertThat(result.read("@.Command", String.class))
                .isEqualTo("GetAuthenticationToken");
        assertThat(result.read("@.PublicKeyECIES", String.class))
                .startsWith("brainpoolP256r1 0x");
        if (request.getCertificateAsX509().getPublicKey() instanceof RSAPublicKey) {
            assertThat(result.read("@.Signature", String.class))
                    .hasSize(344);
        } else {
            assertThat(result.read("@.Signature", String.class))
                    .hasSize(96);
        }
        assertThat(result.read("@.Certificate", String.class))
                .startsWith("MII");
        assertThat(result.read("@.EncryptedMessage", String.class))
                .startsWith("brainpoolP256r1 0x");
    }

    // gemSpec_sgd_epa, A_18026, A_18021
    public void checkGetAuthenticationTokenResponse(EncryptedMessageContainer encryptedMessageContainer) {
        assertThat(encryptedMessageContainer.getStatus())
                .isEqualTo(SgdResponseStatus.OK);
        final String decryptedMessage = new String(encryptedMessageContainer.getEncryptedMessage().decryptMessage(clientEphemeralKeyPair));
        assertThat(decryptedMessage)
                .startsWith("Response " + SgdCryptoUtilities.getInstance().convertToLowerCaseHexWith64ByteLength(clientChallenge) + " ")
                .contains(" AT");
    }
}
